#include <cfaToRgb.h>
#include <interpolatePixels.h>

#include <ImfInputFile.h>
#include <ImfRgbaFile.h>
#include <ImfHeader.h>
#include <ImfChannelList.h>
#include <ImfFrameBuffer.h>
#include <ImfCfaTileAttribute.h>
#include <ImfArray.h>
#include <ImfChromaticities.h>
#include <IexMacros.h>
#include <iostream>


using namespace Iex;
using namespace Imf;
using namespace Imath;
using namespace std;


void
cfaToRgb (const char inFileName[],
          const char outFileName[],
          bool verbose)
{
    CfaTileAttribute::registerAttributeType();

    //
    // Read the input file
    //

    if (verbose)
        cout << "reading file " << inFileName << endl;

    int w;
    int h;
    CfaTile cfaTile;

    typedef map <string, Array2D<half>*> ChannelMap;
    ChannelMap pixels;

    {
        InputFile in (inFileName);
        const Header &hdr = in.header();
        
	Box2i dw = hdr.dataWindow();
	w = dw.max.x - dw.min.x + 1;
	h = dw.max.y - dw.min.y + 1;

        if (const CfaTileAttribute *cfaTileAttr =
                hdr.findTypedAttribute<CfaTileAttribute> ("cfatile"))
        {
            cfaTile = cfaTileAttr->value();
        }
        else
        {
            THROW (InputExc, "File " << inFileName << " "
                             "is not a camera raw file.");
        }

        FrameBuffer fb;

        for (int xTile = 0; xTile < cfaTile.xSize(); ++xTile)
        {
            for (int yTile = 0; yTile < cfaTile.ySize(); ++yTile)
            {
                const CfaPixel &cfap = cfaTile.pixel (xTile, yTile);
                const char *channelName = cfap.channelName.text();

                if (!channelName[0])
                    continue;

                if (const Channel *channel =
                        hdr.channels().findChannel (channelName))
                {
                    if (fb.findSlice (channelName))
                        continue;

                    pixels[channelName] = new Array2D<half> (h, w);
                    Array2D<half> &slicePixels = *pixels[channelName];

                    for (int y = 0; y < h; ++y)
                        for (int x = 0; x < w; ++x)
                            slicePixels[y][x] = half::qNan();

                    size_t xStride =
                        channel->xSampling * sizeof (slicePixels[0][0]);

                    size_t yStride =
                        channel->ySampling * sizeof (slicePixels[0][0]) * w;

                    char *base =
                        (char *)&slicePixels[0][0] -
                        dw.min.x * xStride -
                        dw.min.y * yStride;

                    fb.insert (channelName,
                               Slice (HALF,
                                      base,
                                      xStride,
                                      yStride,
                                      channel->xSampling,
                                      channel->ySampling));

                }
                else
                {
                    THROW (InputExc, "The CFA tile description in file " <<
                                     inFileName << " refers to channel " <<
                                     channelName << ", but the file "
                                     "contains no such channel.");
                }
            }
        }

        in.setFrameBuffer (fb);
        in.readPixels (dw.min.y, dw.max.y);
    }

    //
    // Move the data in our pixel array.
    //

    if (verbose)
        cout << "interpolating" << endl;

    for (int yTile = 0; yTile < cfaTile.ySize(); ++yTile)
    {
        for (int xTile = 0; xTile < cfaTile.xSize(); ++xTile)
        {
            const CfaPixel &cfap = cfaTile.pixel (xTile, yTile);
            const char *channelName = cfap.channelName.text();

            if (!channelName[0])
                continue;

            if (verbose)
            {
                cout << "(" << xTile << ", " << yTile << ") " <<
                        channelName << ", "
                        "xOffset = " << cfap.xOffset << ", "
                        "yOffset = " << cfap.yOffset << ", "
                        "XYZ = " << cfap.XYZ << endl;
            }

            if (xTile + cfap.xOffset < 0 ||
                xTile + cfap.xOffset >= cfaTile.xSize() ||
                yTile + cfap.yOffset < 0 ||
                yTile + cfap.yOffset >= cfaTile.ySize())
            {
                THROW (InputExc, "Invalid pixel offset in CFA tile "
                                 "description attribute in file " <<
                                 inFileName << ".");
            }

            Array2D<half> &slicePixels = *pixels[channelName];

            for (int y = yTile; y < h; y += cfaTile.ySize())
            {
                for (int x = xTile; x < w; x += cfaTile.xSize())
                {
                    half tmp =
                        slicePixels[y + cfap.yOffset][x + cfap.xOffset];

                    slicePixels[y + cfap.yOffset][x + cfap.xOffset] =
                        slicePixels[y][x];

                    slicePixels[y][x] = tmp;
                }
            }
        }
    }

    //
    // Fill all pixels that contain NaNs by interpolating
    // pixels that do not contains NaNs.
    //

    for (ChannelMap::iterator i = pixels.begin(); i != pixels.end(); ++i)
        interpolatePixels (*i->second, w, h);

    //
    // Combine the interpolated channels into CIE XYZ values
    //

    Array2D <Rgba> rgbPixels (h, w);

    for (int y = 0; y < h; ++y)
        for (int x = 0; x < w; ++x)
            rgbPixels[y][x] = Rgba (0, 0, 0, 0);

    for (int xTile = 0; xTile < cfaTile.xSize(); ++xTile)
    {
        for (int yTile = 0; yTile < cfaTile.ySize(); ++yTile)
        {
            const CfaPixel &cfap = cfaTile.pixel (xTile, yTile);
            const char *channelName = cfap.channelName.text();

            ChannelMap::iterator i = pixels.find (channelName); 

            if (i == pixels.end())
                continue;

            const Array2D<half> &slicePixels = *i->second;

            for (int y = 0; y < h; ++y)
            {
                for (int x = 0; x < w; ++x)
                {
                    float p = slicePixels[y][x];
                    Rgba &rgb = rgbPixels[y][x];

                    rgb.r += p * cfap.XYZ.x;
                    rgb.g += p * cfap.XYZ.y;
                    rgb.b += p * cfap.XYZ.z;
                }
            }

            delete i->second;
            pixels.erase (i);
        }
    }

    //
    // Convert the CIE XYZ to RGB values with Rec. 709 / sRGB
    // primaries and white point.
    //

    Chromaticities chr; // default-initialized to Rec. 709
    M44f M = XYZtoRGB (chr, 1);

    for (int y = 0; y < h; ++y)
    {
        for (int x = 0; x < w; ++x)
        {
            Rgba &pixel = rgbPixels[y][x];
            V3f rgb = V3f (pixel.r, pixel.g, pixel.b) * M;
            pixel.r = rgb[0];
            pixel.g = rgb[1];
            pixel.b = rgb[2];
        }
    }

    //
    // Write the output file
    //

    if (verbose)
        cout << "writing file " << outFileName << endl;

    {
        RgbaOutputFile out (outFileName, w, h, WRITE_RGB);
        out.setFrameBuffer (&rgbPixels[0][0], 1, w);
        out.writePixels (h);
    }

    if (verbose)
        cout << "done." << endl;
}
