#include <rgbToCfaRgb.h>
#include <ImfRgbaFile.h>
#include <ImfOutputFile.h>
#include <ImfChannelList.h>
#include <ImfFrameBuffer.h>
#include <ImfCfaTileAttribute.h>
#include <ImfArray.h>
#include <Iex.h>
#include <iostream>
#include <algorithm>


using namespace Iex;
using namespace Imf;
using namespace Imath;
using namespace std;


void
rgbToCfaRgb (const char inFileName[], const char outFileName[], bool verbose)
{
    //
    // Read the input file
    //

    if (verbose)
        cout << "reading file " << inFileName << endl;

    Array2D<Rgba> rgb;
    int w;
    int h;

    {
        RgbaInputFile in (inFileName);
        
	Box2i dw = in.dataWindow();
	w = dw.max.x - dw.min.x + 1;
	h = dw.max.y - dw.min.y + 1;

	rgb.resizeErase (h, w);
	in.setFrameBuffer (&rgb[0][0] - dw.min.x - dw.min.y * w, 1, w);
	in.readPixels (dw.min.y, dw.max.y);
    }

    //
    // Make sure the width and height of the image are even
    //

    if (w & 1)
        --w;

    if (h & 1)
        --h;

    //
    // Low-pass filter the R, G and B channels,
    // first horizontally, then vertically.
    //

    Array2D<Rgba> filteredRgb (h, w);

    for (int y = 0; y < h; ++y)
    {
        for (int x = 0; x < w; ++x)
        {
            const Rgba &p1 = rgb[y][max (x - 1, 0)];
            const Rgba &p2 = rgb[y][x];
            const Rgba &p3 = rgb[y][min (x + 1, w - 1)];

            Rgba &q = filteredRgb[y][x];

            q.r = p1.r * 0.25f + p2.r * 0.50f + p3.r * 0.25f; 
            q.g = p1.g * 0.25f + p2.g * 0.50f + p3.g * 0.25f; 
            q.b = p1.b * 0.25f + p2.b * 0.50f + p3.b * 0.25f;
        }
    }

    for (int y = 0; y < h; ++y)
    {
        for (int x = 0; x < w; ++x)
        {
            const Rgba &p1 = rgb[max (y - 1, 0)][x];
            const Rgba &p2 = rgb[y][x];
            const Rgba &p3 = rgb[min (y + 1, h - 1)][x];

            Rgba &q = filteredRgb[y][x];

            q.r = p1.r * 0.25f + p2.r * 0.50f + p3.r * 0.25f; 
            q.g = p1.g * 0.25f + p2.g * 0.50f + p3.g * 0.25f; 
            q.b = p1.b * 0.25f + p2.b * 0.50f + p3.b * 0.25f;
        }
    }

    //
    // Sub-sample the R, G, and B channels in a Bayer pattern:
    // The original pixel grid
    //
    //      RGB RGB RGB RGB RGB RGB
    //
    //      RGB RGB RGB RGB RGB RGB
    //
    //      RGB RGB RGB RGB RGB RGB
    //
    //      RGB RGB RGB RGB RGB RGB
    //
    //      RGB RGB RGB RGB RGB RGB
    //
    // becomes
    //
    //      G   R   G   R   G   R
    //
    //      B   G   B   G   B   G
    //
    //      G   R   G   R   G   R
    //
    //      B   G   B   G   B   G
    //

    Array2D <half> cfaR (h / 2, w / 2);
    Array2D <half> cfaG (h, w / 2);
    Array2D <half> cfaB (h / 2, w / 2);

    for (int y = 0; y < h; ++y)
    {
        for (int x = 0; x < w; ++x)
        {
            if ((y & 1) == 0 && (x & 1) == 1)
                cfaR[y / 2][x / 2] = filteredRgb[y][x].r;

            if (((x + y) & 1) == 0)
                cfaG[y][x / 2] = filteredRgb[y][x].g;

            if ((y & 1) == 1 && (x & 1) == 0)
                cfaB[y / 2][x / 2] = filteredRgb[y][x].b;
        }
    }

    //
    // Write the output file.
    //

    if (verbose)
        cout << "writing file " << outFileName << endl;

    Header hdr (w, h);
    hdr.compression() = PIZ_COMPRESSION;

    hdr.channels().insert ("cfaR",              // name
                           Channel (HALF,       // type
                                    2,          // xSampling
                                    2));        // ySampling

    hdr.channels().insert ("cfaG",              // name
                           Channel (HALF,       // type
                                    2,          // xSampling
                                    1));        // ySampling

    hdr.channels().insert ("cfaB",              // name
                           Channel (HALF,       // type
                                    2,          // xSampling
                                    2));        // ySampling

    CfaTile cfaTile (2, 2);

    cfaTile.pixel (0, 0) =
        CfaPixel ("cfaG",                       // channelName
                  0,                            // xOffset
                  0,                            // yOffset
                  V3f (0.3576, 0.7152, 0.1192));// XYZ

    cfaTile.pixel (1, 0) =
        CfaPixel ("cfaR",                       // channelName
                  -1,                           // xOffset
                  0,                            // yOffset
                  V3f (0.4124, 0.2126, 0.0193));// XYZ

    cfaTile.pixel (0, 1) =
        CfaPixel ("cfaB",                       // channelName
                  0,                            // xOffset
                  -1,                           // yOffset
                  V3f (0.1805, 0.0722, 0.9505));// XYZ

    cfaTile.pixel (1, 1) =
        CfaPixel ("cfaG",                       // channelName
                  -1,                           // xOffset
                  0,                            // yOffset
                  V3f (0.3576, 0.7152, 0.1192));// XYZ

    hdr.insert ("cfatile", CfaTileAttribute (cfaTile));

    FrameBuffer fb;

    fb.insert ("cfaR",                          // name
               Slice (HALF,                     // type
                      (char *)&cfaR[0][0],      // base
                      sizeof (cfaR[0][0]),      // xStride
                      sizeof (cfaR[0][0]) * w / 2,  // yStride
                      2,                        // xSampling
                      2));                      // ySampling

    fb.insert ("cfaG",                          // name
               Slice (HALF,                     // type
                      (char *)&cfaG[0][0],      // base
                      sizeof (cfaG[0][0]),      // xStride
                      sizeof (cfaG[0][0]) * w / 2,  // yStride
                      2,                        // xSampling
                      1));                      // ySampling

    fb.insert ("cfaB",                          // name
               Slice (HALF,                     // type
                      (char *)&cfaB[0][0],      // base
                      sizeof (cfaB[0][0]),      // xStride
                      sizeof (cfaB[0][0]) * w / 2,  // yStride
                      2,                        // xSampling
                      2));                      // ySampling

    {
        OutputFile out (outFileName, hdr);
        out.setFrameBuffer (fb);
        out.writePixels (h);
    }

    if (verbose)
        cout << "done." << endl;
}
